### api_table.py ###

from datetime import datetime
from typing import cast

import sqlalchemy
from sqlalchemy.orm import declarative_base
from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.media_model import (
    TemplateMediaModel,
)

LegacyBase = declarative_base()


class ApiModel:
    __tablename__ = ""
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    post_id = cast(
        int, sqlalchemy.Column(sqlalchemy.Integer, unique=True, nullable=False)
    )
    text = cast(str, sqlalchemy.Column(sqlalchemy.String))
    price = cast(int, sqlalchemy.Column(sqlalchemy.Integer))
    paid = sqlalchemy.Column(sqlalchemy.Integer)
    archived = cast(bool, sqlalchemy.Column(sqlalchemy.Boolean, default=False))
    created_at = cast(datetime, sqlalchemy.Column(sqlalchemy.TIMESTAMP))
    medias: list[TemplateMediaModel] = []

    def legacy(self, table_name: str):
        class legacy_api_table(LegacyBase):
            __tablename__ = table_name
            id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
            text = sqlalchemy.Column(sqlalchemy.String)
            price = sqlalchemy.Column(sqlalchemy.Integer)
            paid = sqlalchemy.Column(sqlalchemy.Integer)
            created_at = sqlalchemy.Column(sqlalchemy.DATETIME)

        return legacy_api_table

    def convert(self):
        item = self.__dict__
        item.pop("_sa_instance_state")
        return item

    def find_media(self, media_id: int):
        for db_media in self.medias:
            if db_media.media_id == media_id:
                return db_media



from __future__ import annotations

import copy
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Literal

import ultima_scraper_api
from aiohttp.client_reqrep import ClientResponse
from ultima_scraper_api.classes.prepare_directories import FormatTypes
from ultima_scraper_api.helpers.main_helper import open_partial
from ultima_scraper_api.managers.session_manager import EXCEPTION_TEMPLATE
from ultima_scraper_renamer.reformat import ReformatItem

if TYPE_CHECKING:
    api_types = ultima_scraper_api.api_types
    user_types = ultima_scraper_api.user_types


class FilesystemManager:
    def __init__(self) -> None:
        self.user_data_directory = Path("__user_data__")
        self.trash_directory = self.user_data_directory.joinpath("trash")
        self.profiles_directory = self.user_data_directory.joinpath("profiles")
        self.devices_directory = self.user_data_directory.joinpath("devices")
        self.settings_directory = Path("__settings__")
        self.ignore_files = ["desktop.ini", ".DS_Store", ".DS_store", "@eaDir"]
        self.directory_manager: DirectoryManager | None = None
        self.directory_manager_users: dict[int, DirectoryManager] = {}
        self.file_manager_users: dict[int, FileManager] = {}

    def __iter__(self):
        for each in self.__dict__.values():
            yield each

    def check(self):
        for directory in self:
            if isinstance(directory, Path):
                directory.mkdir(exist_ok=True)

    def move(self, src: Path, trg: Path):
        shutil.move(src, trg)

    def remove_mandatory_files(
        self, files: list[Path] | Generator[Path, None, None], keep: list[str] = []
    ):
        folders = [x for x in files if x.name not in self.ignore_files]
        if keep:
            folders = [x for x in files if x.name in keep]
        return folders

    def get_directory_manager(self, user_id: int):
        return self.directory_manager_users[user_id]

    def get_file_manager(self, user_id: int):
        return self.file_manager_users[user_id]

    def activate_directory_manager(self, api: api_types):
        from ultima_scraper_api.helpers import main_helper

        site_settings = api.get_site_settings()
        root_metadata_directory = main_helper.check_space(
            site_settings.metadata_directories
        )
        root_download_directory = main_helper.check_space(
            site_settings.download_directories
        )
        self.directory_manager = DirectoryManager(
            site_settings,
            root_metadata_directory,
            root_download_directory,
        )

    def trash(self):
        pass

    async def write_data(
        self, response: ClientResponse, download_path: Path, callback: Any = None
    ):
        status_code = None
        if response.status == 200:
            total_length = 0
            os.makedirs(os.path.dirname(download_path), exist_ok=True)
            partial_path: str | None = None
            try:
                with open_partial(download_path) as f:
                    partial_path = f.name
                    try:
                        async for data in response.content.iter_chunked(4096):
                            f.write(data)
                            length = len(data)
                            total_length += length
                            if callback:
                                callback(length)
                    except EXCEPTION_TEMPLATE as _e:
                        status_code = 1
                    except Exception as _e:
                        raise Exception(f"Unknown Error: {_e}")
            except:
                if partial_path:
                    os.unlink(partial_path)
                raise
            else:
                if status_code:
                    os.unlink(partial_path)
                else:
                    try:
                        os.replace(partial_path, download_path)
                    except OSError:
                        pass
        else:
            if response.content_length:
                pass
                # progress_bar.update_total_size(-response.content_length)
            status_code = 2
        return status_code

    async def create_directory_manager(self, api: api_types, user: user_types):
        # profile_directory = filesystem_directory_manager.profile.root_directory.joinpath(
        #     self.username
        # )
        if self.directory_manager:
            directory_manager = DirectoryManager(
                api.get_site_settings(),
                self.directory_manager.root_metadata_directory,
                self.directory_manager.root_download_directory,
            )
            option = {
                "site_name": api.site_name,
                "profile_username": user.username,
                "model_username": user.username,
                "directory": directory_manager.root_download_directory,
            }
            reformat_item_fd = ReformatItem(option)
            _f_d_p = reformat_item_fd.remove_non_unique(
                directory_manager, "file_directory_format"
            )
            # f_d_p.mkdir(parents=True, exist_ok=True)
            option["directory"] = self.directory_manager.root_metadata_directory
            reformat_item_md = ReformatItem(option)
            _f_d_p_2 = reformat_item_md.remove_non_unique(
                self.directory_manager, "metadata_directory_format"
            )
            # f_d_p_2.mkdir(parents=True, exist_ok=True)
            self.directory_manager_users[user.id] = directory_manager
            self.file_manager_users[user.id] = FileManager(directory_manager)
            return directory_manager

    async def format_directories(self, subscription: user_types) -> DirectoryManager:
        directory_manager = self.get_directory_manager(subscription.id)
        file_manager = self.get_file_manager(subscription.id)
        authed = subscription.get_authed()
        api = authed.api
        site_settings = authed.api.get_site_settings()
        if site_settings:
            authed_username = authed.username
            subscription_username = subscription.username
            site_name = authed.api.site_name
            reformat_item = ReformatItem()
            reformat_item.site_name = site_name
            reformat_item.profile_username = authed_username
            reformat_item.model_username = subscription_username
            reformat_item.date = datetime.today()
            reformat_item.date_format = site_settings.date_format
            reformat_item.text_length = site_settings.text_length
            reformat_item.directory = directory_manager.root_metadata_directory
            string = reformat_item.reformat(site_settings.metadata_directory_format)
            directory_manager.user.metadata_directory = Path(string)
            reformat_item_2 = copy.copy(reformat_item)
            reformat_item_2.directory = directory_manager.root_download_directory
            string = reformat_item_2.reformat(site_settings.file_directory_format)
            formtatted_root_download_directory = reformat_item_2.remove_non_unique(
                directory_manager, "file_directory_format"
            )
            directory_manager.user.download_directory = (
                formtatted_root_download_directory
            )
            await file_manager.set_default_files(reformat_item, reformat_item_2)
            _metadata_filepaths = await file_manager.find_metadata_files(
                legacy_files=False
            )
            # I forgot why we need to set default file twice
            await file_manager.set_default_files(reformat_item, reformat_item_2)
            user_metadata_directory = directory_manager.user.metadata_directory
            _user_download_directory = directory_manager.user.download_directory
            legacy_metadata_directory = user_metadata_directory
            directory_manager.user.legacy_metadata_directories.append(
                legacy_metadata_directory
            )
            items = api.ContentTypes().__dict__.items()
            for api_type, _ in items:
                legacy_metadata_directory_2 = user_metadata_directory.joinpath(api_type)
                directory_manager.user.legacy_metadata_directories.append(
                    legacy_metadata_directory_2
                )
            legacy_model_directory = directory_manager.root_download_directory.joinpath(
                site_name, subscription_username
            )
            directory_manager.user.legacy_download_directories.append(
                legacy_model_directory
            )
        return directory_manager


from ultima_scraper_api.classes.make_settings import SiteSettings


class DirectoryManager:
    def __init__(
        self,
        site_settings: SiteSettings,
        root_metadata_directory: Path = Path(),
        root_download_directory: Path = Path(),
    ) -> None:
        self.root_directory = Path()
        self.root_metadata_directory = Path(root_metadata_directory)
        self.root_download_directory = Path(root_download_directory)
        # self.profile = self.ProfileDirectories(Path(profile_directory))
        self.user = self.UserDirectories()
        formats = FormatTypes(site_settings)
        string, status = formats.check_rules()
        if not status:
            print(string)
            exit(0)
        self.formats = formats
        pass

    def create_directories(self):
        # self.profile.create_directories()
        self.root_metadata_directory.mkdir(exist_ok=True)
        self.root_download_directory.mkdir(exist_ok=True)

    def delete_empty_directories(
        self, directory: Path, filesystem_manager: FilesystemManager
    ):
        for root, dirnames, _files in os.walk(directory, topdown=False):
            for dirname in dirnames:
                full_path = os.path.realpath(os.path.join(root, dirname))
                contents = os.listdir(full_path)
                if not contents:
                    shutil.rmtree(full_path, ignore_errors=True)
                else:
                    content_paths = [Path(full_path, content) for content in contents]
                    contents = filesystem_manager.remove_mandatory_files(content_paths)
                    if not contents:
                        shutil.rmtree(full_path, ignore_errors=True)

        if os.path.exists(directory) and not os.listdir(directory):
            os.rmdir(directory)

    # class ProfileDirectories:
    #     def __init__(self, root_directory: Path) -> None:
    #         self.root_directory = Path(root_directory)
    #         self.metadata_directory = self.root_directory.joinpath("Metadata")
    #     def create_directories(self):
    #         self.root_directory.mkdir(exist_ok=True)

    class UserDirectories:
        def __init__(self) -> None:
            self.metadata_directory = Path()
            self.download_directory = Path()
            self.legacy_download_directories: list[Path] = []
            self.legacy_metadata_directories: list[Path] = []

        def find_legacy_directory(
            self,
            directory_type: Literal["metadata", "download"] = "metadata",
            api_type: str = "",
        ):
            match directory_type:
                case "metadata":
                    directories = self.legacy_metadata_directories
                case _:
                    directories = self.legacy_download_directories
            final_directory = directories[0]
            for directory in directories:
                for part in directory.parts:
                    if api_type in part:
                        return directory
            return final_directory

    async def walk(self, directory: Path):
        all_files: list[Path] = []
        for root, _subdirs, files in os.walk(directory):
            x = [Path(root, x) for x in files]
            all_files.extend(x)
        return all_files


class FileManager:
    def __init__(self, directory_manager: DirectoryManager) -> None:
        self.files: list[Path] = []
        self.directory_manager = directory_manager

    async def set_default_files(
        self,
        prepared_metadata_format: ReformatItem,
        prepared_download_format: ReformatItem,
    ):
        self.files = []
        await self.add_files(prepared_metadata_format, "metadata_directory_format")
        await self.add_files(prepared_download_format, "file_directory_format")

    async def add_files(self, reformatter: ReformatItem, format_key: str):
        directory_manager = self.directory_manager
        formatted_directory = reformatter.remove_non_unique(
            directory_manager, format_key
        )
        files: list[Path] = []
        files = await directory_manager.walk(formatted_directory)
        self.files.extend(files)
        return files

    async def find_metadata_files(self, legacy_files: bool = True):
        new_list: list[Path] = []
        for filepath in self.files:
            if not legacy_files:
                if "__legacy_metadata__" in filepath.parts:
                    continue
            match filepath.suffix:
                case ".db":
                    red_list = ["thumbs.db"]
                    status = [x for x in red_list if x == filepath.name.lower()]
                    if status:
                        continue
                    new_list.append(filepath)
                case ".json":
                    new_list.append(filepath)
                case _:
                    pass
        return new_list



from __future__ import annotations

import copy
from itertools import product
from typing import Any, Optional

import ultima_scraper_api
import ultima_scraper_api.classes.make_settings as make_settings
from sqlalchemy.exc import OperationalError
from tqdm.asyncio import tqdm_asyncio
from ultima_scraper_renamer import renamer

from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.api_model import (
    ApiModel,
)
from ultima_scraper_collection.managers.database_manager.connections.sqlite.sqlite_database import (
    DBCollection,
)
from ultima_scraper_collection.managers.database_manager.database_manager import (
    DatabaseManager,
)
from ultima_scraper_collection.managers.download_manager import DownloadManager
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    MetadataManager,
)

auth_types = ultima_scraper_api.auth_types
user_types = ultima_scraper_api.user_types
message_types = ultima_scraper_api.message_types
error_types = ultima_scraper_api.error_types
subscription_types = ultima_scraper_api.subscription_types

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ultima_scraper_collection.managers.datascraper_manager.datascrapers.fansly import (
        FanslyDataScraper,
    )
    from ultima_scraper_collection.managers.datascraper_manager.datascrapers.onlyfans import (
        OnlyFansDataScraper,
    )
    from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
        ContentMetadata,
        MediaMetadata,
    )

    datascraper_types = OnlyFansDataScraper | FanslyDataScraper


class download_session(tqdm_asyncio):
    def start(
        self,
        unit: str = "B",
        unit_scale: bool = True,
        miniters: int = 1,
        tsize: int = 0,
    ):
        self.unit = unit
        self.unit_scale = unit_scale
        self.miniters = miniters
        self.total = 0
        self.colour = "Green"
        if tsize:
            tsize = int(tsize)
            self.total += tsize

    def update_total_size(self, tsize: Optional[int]):
        if tsize:
            tsize = int(tsize)
            self.total += tsize


class StreamlinedDatascraper:
    def __init__(self, datascraper: datascraper_types) -> None:
        self.datascraper = datascraper
        self.filesystem_manager = FilesystemManager()
        self.content_types = self.datascraper.api.ContentTypes()
        self.media_types = self.datascraper.api.MediaTypes()
        self.user_list: set[user_types] = set()
        self.metadata_manager_users: dict[int, MetadataManager] = {}

    async def configure_datascraper_jobs(self):
        api = self.datascraper.api
        site_settings = api.get_site_settings()
        available_jobs = site_settings.jobs.scrape
        option_manager = self.datascraper.option_manager
        performer_options = option_manager.performer_options
        valid_user_list: set[user_types] = set(
            option_manager.subscription_options.final_choices
        )
        scraping_subscriptions = (
            self.datascraper.api.get_site_settings().jobs.scrape.subscriptions
        )
        identifiers = []
        if performer_options:
            identifiers = performer_options.return_auto_choice()
        if not available_jobs.subscriptions:
            for authed in api.auths:
                authed.subscriptions = []
        if available_jobs.messages:
            chat_users: list[user_types] = []
            if identifiers:
                for authed in api.auths:
                    for identifier in identifiers:
                        chat_user = await authed.get_user(identifier)
                        if isinstance(chat_user, user_types):
                            chat_users.append(chat_user)
            else:
                chat_users = await self.get_chat_users()
            [
                user.scrape_whitelist.append("Messages")
                for user in chat_users
                if not await user.is_subscribed() or not scraping_subscriptions
            ]
            [valid_user_list.add(x) for x in chat_users]

        if available_jobs.paid_contents:
            for authed in self.datascraper.api.auths:
                paid_contents = await authed.get_paid_content()
                if not isinstance(paid_contents, error_types):
                    for paid_content in paid_contents:
                        author = paid_content.get_author()
                        if identifiers:
                            found = await author.match_identifiers(identifiers)
                            if not found:
                                continue
                        if author:
                            performer = authed.find_user_by_identifier(
                                identifier=author.id,
                            )
                            if performer:
                                performer.job_whitelist.append("PaidContents")
                                performer.scrape_whitelist.clear()
                                valid_user_list.add(performer)
        # Need to filter out own profile with is_performer,etc
        final_valid_user_set = {
            user
            for user in valid_user_list
            if user.username not in user.get_authed().blacklist
        }

        self.user_list = final_valid_user_set
        return final_valid_user_set

    # Prepares the API links to be scraped

    async def prepare_scraper(
        self,
        user: user_types,
        metadata_manager: MetadataManager,
        content_type: str,
        master_set: list[Any] = [],
    ):
        authed = user.get_authed()
        current_job = user.get_current_job()
        if not current_job:
            return
        temp_master_set: list[Any] = copy.copy(master_set)
        if not temp_master_set:
            match content_type:
                case "Stories":
                    temp_master_set.extend(await self.datascraper.get_all_stories(user))
                    pass
                case "Posts":
                    temp_master_set = await self.datascraper.get_all_posts(user)
                case "Messages":
                    pass
                    unrefined_set: list[message_types | Any] = await user.get_messages()
                    mass_messages = getattr(authed, "mass_messages")
                    if user.is_me() and mass_messages:
                        mass_messages = getattr(authed, "mass_messages")
                        # Need access to a creator's account to fix this
                        # unrefined_set2 = await self.datascraper.process_mass_messages(
                        #     authed,
                        #     mass_messages,
                        # )
                        # unrefined_set += unrefined_set2
                    temp_master_set = unrefined_set
                case "Archived":
                    pass
                case "Chats":
                    pass
                case "Highlights":
                    pass
                case "MassMessages":
                    pass
                case _:
                    raise Exception(f"{content_type} is an invalid choice")
        # Adding paid content and removing duplicates by id
        if isinstance(user, ultima_scraper_api.onlyfans_classes.user_model.create_user):
            for paid_content in await user.get_paid_contents(content_type):
                temp_master_set.append(paid_content)
            temp_master_set = list({getattr(obj,"id"): obj for obj in temp_master_set}.values())
        await self.process_scraped_content(
            temp_master_set, content_type, user, metadata_manager
        )
        current_job.done = True

    async def process_scraped_content(
        self,
        master_set: list[dict[str, Any]],
        api_type: str,
        subscription: user_types,
        metadata_manager: MetadataManager,
    ):
        if not master_set:
            return False
        # if  api_type != "Posts":
        #     return
        authed = subscription.get_authed()
        subscription_directory_manager = self.filesystem_manager.get_directory_manager(
            subscription.id
        )
        formatted_metadata_directory = (
            subscription_directory_manager.user.metadata_directory
        )
        unrefined_set = []
        with authed.get_pool() as pool:
            print(f"Processing Scraped {api_type}")
            tasks = pool.starmap(
                self.datascraper.media_scraper,
                product(
                    master_set,
                    [subscription],
                    [subscription_directory_manager.root_download_directory],
                    [api_type],
                ),
            )
            settings = {"colour": "MAGENTA"}
            unrefined_set: list[dict[str, Any]] = await tqdm_asyncio.gather(
                *tasks, **settings
            )
            new_metadata = metadata_manager.merge_content_and_directories(unrefined_set)
            final_content, _final_directories = new_metadata
            metadata_path = formatted_metadata_directory.joinpath("user_data.db")
            metadata_manager.db_manager = DatabaseManager().get_sqlite_db(metadata_path)
            if new_metadata:
                new_metadata_content = final_content
                metadata_manager.content_metadatas.extend(new_metadata_content)
                subscription.scrape_manager.set_scraped(api_type, new_metadata_content)
                if new_metadata_content:
                    import_status = metadata_manager.db_manager.import_metadata(
                        new_metadata_content, api_type
                    )
                    if import_status:
                        Session = metadata_manager.db_manager.session_factory
                        if authed.api.config.settings.helpers.renamer:
                            print(f"{subscription.username}: Renaming files.")
                            _new_metadata_object = await renamer.start(
                                subscription,
                                subscription_directory_manager,
                                api_type,
                                Session,
                            )
                        pass
            else:
                print(f"No {api_type} found.")
        return True

    # Downloads scraped content

    async def prepare_downloads(self, performer: user_types, api_type: str):
        metadata_manager = self.metadata_manager_users[performer.id]
        global_settings = performer.get_api().get_global_settings()
        site_settings = performer.get_api().get_site_settings()
        if not (global_settings and site_settings):
            return
        filesystem_manager = self.datascraper.filesystem_manager
        performer_directory_manager = filesystem_manager.get_directory_manager(
            performer.id
        )
        directory = performer_directory_manager.user.download_directory
        current_job = performer.get_current_job()

        metadata_path = performer_directory_manager.user.metadata_directory.joinpath(
            "user_data.db"
        )
        user_data_db = DatabaseManager().get_sqlite_db(metadata_path)
        if user_data_db and user_data_db.name.exists():
            database_session = user_data_db.session
            db_collection = DBCollection()
            database = db_collection.database_picker("user_data")
            api_table = database.table_picker(api_type)
            media_table = database.media_table
            overwrite_files = site_settings.overwrite_files
            final_download_set: set[ContentMetadata] = set()
            final_media_set: set[MediaMetadata] = set()
            total_media_count = 0
            if database:
                db_content_dict = {}
                db_posts: list[ApiModel] = database_session.query(api_table)
                for db_post in db_posts:
                    if db_post.post_id not in db_content_dict:
                        db_content_dict[db_post.post_id] = db_post
                    else:
                        raise Exception("Duplicate key in db_content_dict")

                content_metadatas = metadata_manager.content_metadatas
                for content_metadata in content_metadatas:
                    if content_metadata.api_type == api_type:
                        db_post = db_content_dict[content_metadata.content_id]
                        if content_metadata.content_id == db_post.post_id:
                            content_metadata.__db_content__ = db_post
                            db_media_query = (
                                database_session.query(media_table)
                                .filter(media_table.post_id == db_post.post_id)
                                .filter(media_table.api_type == api_type)
                            )
                            if overwrite_files:
                                db_post.medias = db_media_query.all()
                            else:
                                db_post.medias = db_media_query.filter(
                                    media_table.downloaded == False
                                ).all()
                            if db_post.medias:
                                final_download_set.add(content_metadata)
                                [
                                    final_media_set.add(x.id)
                                    for x in content_metadata.medias
                                ]
                                total_media_count += len(content_metadata.medias)
                download_manager = DownloadManager(
                    filesystem_manager,
                    performer.get_session_manager(),
                    final_download_set,
                    global_settings.helpers.reformat_media,
                )
                download_count = len(final_media_set)
                duplicate_count = total_media_count - download_count
                string = "Processing Download:\n"
                string += f"Name: {performer.username} | Type: {api_type} | Downloading: {download_count} | Total: {total_media_count} | Duplicates: {duplicate_count} | Directory: {directory}\n"
                if total_media_count:
                    print(string)
                    _result = await download_manager.bulk_download()
                    pass
                while True:
                    try:
                        database_session.commit()
                        break
                    except OperationalError:
                        database_session.rollback()
                database_session.close()
            current_job.done = True

    async def manage_subscriptions(
        self,
        authed: auth_types,
        identifiers: list[int | str] = [],
        refresh: bool = True,
    ):
        temp_subscriptions: list[subscription_types] = []
        results = await self.datascraper.get_all_subscriptions(
            authed, identifiers, refresh
        )
        site_settings = authed.api.get_site_settings()
        if not site_settings:
            return temp_subscriptions
        ignore_type = site_settings.ignore_type
        results.sort(key=lambda x: x.user.is_me(), reverse=True)
        for result in results:
            # await result.create_directory_manager(user=True)
            subscribe_price = result.get_price()
            if ignore_type in ["paid"]:
                if subscribe_price > 0:
                    continue
            if ignore_type in ["free"]:
                if subscribe_price == 0:
                    continue
            temp_subscriptions.append(result)
        authed.subscriptions = temp_subscriptions
        return authed.subscriptions

    async def account_setup(
        self,
        auth: auth_types,
        datascraper: datascraper_types,
        site_settings: make_settings.SiteSettings,
        identifiers: list[int | str] | list[str] = [],
    ) -> tuple[bool, list[user_types]]:
        status = False
        subscriptions: list[subscription_types] = []
        auth = await auth.login()

        if auth.check_authed() and site_settings:
            authed = auth
            # metadata_filepath = (
            #     authed.directory_manager.profile.metadata_directory.joinpath(
            #         "Mass Messages.json"
            #     )
            # )
            # if authed.isPerformer:
            #     imported = main_helper.import_json(metadata_filepath)
            #     if "auth" in imported:
            #         imported = imported["auth"]
            #     mass_messages = await authed.get_mass_messages(resume=imported)
            #     if mass_messages:
            #         main_helper.export_json(mass_messages, metadata_filepath)
            authed.blacklist = await authed.get_blacklist(
                authed.api.get_site_settings().blacklists
            )
            if identifiers or site_settings.jobs.scrape.subscriptions:
                subscriptions.extend(
                    await datascraper.manage_subscriptions(
                        authed, identifiers=identifiers  # type: ignore
                    )
                )
            status = True
        elif (
            auth.auth_details.email
            and auth.auth_details.password
            and site_settings.browser.auth
        ):
            # domain = "https://onlyfans.com"
            # oflogin.login(auth, domain, auth.session_manager.get_proxy())
            pass
        return status, subscriptions

    async def get_chat_users(self):
        chat_users: list[user_types] = []
        for authed in self.datascraper.api.auths:
            chats = await authed.get_chats()
            for chat in chats:
                username: str = chat["withUser"].username
                subscription = await authed.get_subscription(identifier=username)
                if not subscription:
                    subscription = chat["withUser"]
                    chat_users.append(subscription)
        return chat_users



from datetime import datetime
from pathlib import Path
from typing import Any
from urllib.parse import ParseResult, urlparse

import ultima_scraper_api
from sqlalchemy import inspect

from ultima_scraper_collection.managers.database_manager.connections.sqlite.sqlite_database import (
    DBCollection,
    SqliteDatabase,
)
from ultima_scraper_collection.managers.database_manager.database_manager import (
    DatabaseManager,
)
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager

api_types = ultima_scraper_api.api_types
user_types = ultima_scraper_api.user_types
content_types = ultima_scraper_api.content_types
from ultima_scraper_api.classes.prepare_metadata import format_content
from ultima_scraper_api.helpers import main_helper

from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.api_model import (
    ApiModel,
)
from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.user_database import (
    messages_table,
)


class MetadataExtractor:
    def __init__(self, post_item: format_content.post_item) -> None:
        self.item = post_item

    def get_id(self):
        return self.item.post_id

    def get_user_id(self):
        return None

    def get_text(self):
        return self.item.text

    def get_preview_ids(self):
        value: list[int] = []
        return value

    def resolve_archived(self):
        return False

    def get_date(self):
        return self.item.createdAt

    def get_medias(self, content_metadata: "ContentMetadata"):
        final_assets: list[MediaMetadata] = []
        for media in self.item.medias:
            new_asset = MediaMetadata(
                media.media_id,
                media.media_type,
                content_metadata,
                created_at=content_metadata.created_at_string,
            )
            new_asset.urls = media.links
            new_asset.directory = media.directory
            new_asset.filename = media.filename
            new_asset.size = media.size
            final_assets.append(new_asset)
        return final_assets

    def resolve_paid(self):
        return self.item.paid

    def get_receiver_id(self):
        if isinstance(self.item, messages_table):
            return


class Extractor:
    def __init__(self, item: content_types) -> None:
        self.item = item

    def get_id(self):
        return self.item.id

    def get_user_id(self):
        return self.item.get_author().id

    def get_text(self):
        if isinstance(self.item, ultima_scraper_api.post_types):
            final_text = self.item.rawText or self.item.text
        elif isinstance(self.item, ultima_scraper_api.message_types):
            final_text = self.item.text
        else:
            final_text = ""
        return final_text

    def get_preview_ids(self):
        if isinstance(self.item, ultima_scraper_api.post_types):
            final_preview_ids = self.item.preview or self.item.preview_ids
        elif isinstance(self.item, ultima_scraper_api.message_types):
            final_preview_ids = self.item.previews
        else:
            final_preview_ids = []
        return final_preview_ids

    def get_date(self):
        temp_date = self.item.createdAt
        default_date = datetime.strftime(datetime.utcnow(), "%d-%m-%Y %H:%M:%S")
        if temp_date == "-001-11-30T00:00:00+00:00":
            final_date_string = default_date
            date_object = datetime.strptime(default_date, "%d-%m-%Y %H:%M:%S")
        else:
            if not temp_date:
                temp_date = default_date
            if isinstance(temp_date, int):
                timestamp = float(temp_date)
                date_object = datetime.fromtimestamp(timestamp)
            else:
                date_object = datetime.fromisoformat(temp_date)
            final_date_string = date_object.replace(tzinfo=None).strftime(
                "%d-%m-%Y %H:%M:%S"
            )
        return final_date_string

    def get_medias(self, content_metadata: "ContentMetadata"):
        final_assets: list[MediaMetadata] = []
        for asset_metadata in self.item.media:
            raw_media_type = asset_metadata["type"]
            if "mimetype" in asset_metadata:
                raw_media_type: str = asset_metadata["mimetype"].split("/")[0]
            author = self.item.get_author()
            media_type = author.get_api().MediaTypes().find_by_value(raw_media_type)
            new_asset = MediaMetadata(
                asset_metadata["id"],
                media_type,
                content_metadata,
                created_at=content_metadata.created_at_string,
            )
            main_url = self.item.url_picker(asset_metadata)
            preview_url = self.item.preview_url_picker(asset_metadata)
            authed = author.get_authed()
            if authed.drm:
                new_asset.drm = bool(authed.drm.has_drm(asset_metadata))
            new_asset.urls = []
            matches = ["us", "uk", "ca", "ca2", "de"]
            for url in [main_url, preview_url].copy():
                if url:
                    if url.hostname:
                        subdomain = url.hostname.split(".")[1]
                        if any(subdomain in nm for nm in matches):
                            subdomain = url.hostname.split(".")[1]
                            if "upload" in subdomain:
                                continue
                            if "convert" in subdomain:
                                continue
                    new_asset.urls.append(url.geturl())

            if int(new_asset.id) in content_metadata.preview_media_ids:
                new_asset.preview = True
            new_asset.__raw__ = asset_metadata
            final_assets.append(new_asset)
        return final_assets

    def resolve_archived(self):
        archived = False
        if isinstance(self.item, ultima_scraper_api.post_types):
            archived = self.item.isArchived
        return archived

    def resolve_paid(self):
        if getattr(self.item, "price", 0):
            if all(media["canView"] for media in self.item.media):
                return True
        return False

    def get_receiver_id(self):
        if isinstance(self.item, ultima_scraper_api.message_types):
            return self.item.get_receiver().id


class ContentMetadata:
    def __init__(self, content_id: int, api_type: str) -> None:
        self.content_id = content_id
        self.user_id: int | None = None
        self.receiver_id: int | None = None
        self.text: str = ""
        self.preview_media_ids: list[int] | list[dict[str, Any]] = []
        self.archived: bool = False
        self.created_at_string: str = ""
        self.medias: list[MediaMetadata] = []
        self.api_type = api_type
        self.price: float | None = None
        self.paid: bool = False
        self.deleted: bool = False
        self.__raw__: Any | None = None
        self.__soft__: Any = None
        self.__db_content__: ApiModel | None = None
        self.__legacy__ = False

    def resolve_extractor(self, result: Extractor | MetadataExtractor):
        self.content_id = result.get_id()
        self.user_id = result.get_user_id()
        self.receiver_id = result.get_receiver_id()
        self.text = result.get_text()
        self.preview_media_ids = result.get_preview_ids()
        self.archived = result.resolve_archived()
        self.created_at_string = result.get_date()
        self.medias: list[MediaMetadata] = result.get_medias(self)
        self.price = getattr(result.item, "price", 0) or 0
        self.paid = result.resolve_paid()
        self.deleted = False
        self.__raw__: Any | None = None
        self.__soft__ = result.item

    def add_media(
        self, media_id: int, media_type: str, urls: list[str], preview: bool = False
    ):
        urls = [url for url in urls if url]
        new_media = MediaMetadata(media_id, media_type, urls, preview, self.created_at)
        self.medias.append(new_media)
        return new_media

    def find_media(self, media_id: int | None = None, urls: list[str] = []):
        for asset in self.medias:
            if asset.id == media_id:
                return asset
        for asset in self.medias:
            for url in urls:
                url = urlparse(url)
                found_url = asset.find_by_url(url)
                if found_url:
                    return asset

    def find_media_metadata_by_id(self, media_id: int):
        for asset in self.medias:
            if asset.id == media_id:
                return asset

    def find_media_by_url(self, urls: list[str]):
        for asset in self.medias:
            for url in urls:
                url = urlparse(url)
                found_url = asset.find_by_url(url)
                if found_url:
                    return asset


class MediaMetadata:
    def __init__(
        self,
        media_id: int,
        media_type: str,
        content_metadata: ContentMetadata,
        urls: list[str] = [],
        preview: bool = False,
        created_at: str = "",
        drm: bool = False,
    ) -> None:
        self.id = int(media_id) if media_id is not None else None
        self.media_type = media_type
        self.urls: list[str] = urls
        self.preview = preview
        self.directory: Path | None = None
        self.filename: str | None = None
        self.size = 0
        self.linked = None
        self.drm = drm
        self.key: str = ""
        self.created_at = created_at or content_metadata.created_at_string
        self.__raw__: Any | None = None
        self.__content_metadata__ = content_metadata

    def find_by_url(self, url: ParseResult):
        for media_url in self.urls:
            media_url = urlparse(media_url)
            url_path = url.path
            if "Protected" in url.path:
                index = url.path.index("/files")
                url_path = url.path[index:]
            if media_url.path == url_path:
                return self

    def get_filepath(self):
        assert self.directory and self.filename
        return Path(self.directory, self.filename)


class MetadataManager:
    def __init__(
        self,
        subscription: user_types,
        filesystem_manager: FilesystemManager,
        db_manager: SqliteDatabase | None = None,
    ) -> None:
        self.subscription = subscription
        self.filesystem_manager = filesystem_manager
        self.db_manager = db_manager
        self.metadatas = self.find_metadatas()
        self.update_metadata_filepaths()
        self.redundant_metadatas: list[Path] = []
        self.content_metadatas: list[ContentMetadata] = []
        self.legacy_content_metadatas: list[ContentMetadata] = []

    def merge_content_and_directories(self, unformatted_set: list[dict[str, Any]]):
        temp_content: set[ContentMetadata] = set()
        temp_directories: set[Path] = set()
        for item in unformatted_set:
            if item:
                temp_content.update(item["content"])
                temp_directories.update(item["directories"])
        final_content = list(temp_content)
        final_directories = list(temp_directories)
        final_content.sort(key=lambda obj: obj.content_id)
        return final_content, final_directories

    def find_metadatas(self):
        found_metadatas: list[Path] = []
        for file_ in self.filesystem_manager.get_file_manager(
            self.subscription.id
        ).files:
            if file_.suffix in [".db", ".json"]:
                found_metadatas.append(file_)
        return found_metadatas

    async def process_legacy_metadata(self):
        # Ignore every other metadata fixer, THIS is the one that will merge all legacy metadata (.json and .db) into user_data.
        self.fix_json()
        # Needs work
        # self.fix_sqlite()
        performer_file_manager = self.filesystem_manager.get_file_manager(
            self.subscription.id
        )
        performer_directory_manager = performer_file_manager.directory_manager
        db_manager = DatabaseManager().get_sqlite_db(
            performer_directory_manager.user.metadata_directory.joinpath("user_data.db")
        )
        db_manager.import_metadata(self.legacy_content_metadatas)
        for legacy_filepath in self.metadatas:
            if (
                legacy_filepath.suffix != ".json"
                or "__legacy__" in legacy_filepath.parts
            ):
                continue
            if legacy_filepath and legacy_filepath.exists():
                new_metadata_filepath = (
                    performer_directory_manager.user.metadata_directory.joinpath(
                        "__legacy__", legacy_filepath.name
                    )
                )
                new_metadata_filepath.parent.mkdir(parents=True, exist_ok=True)
                self.filesystem_manager.move(legacy_filepath, new_metadata_filepath)
        await self.filesystem_manager.format_directories(self.subscription)
        await self.fix_archived_db()
        self.metadatas = self.find_metadatas()

    def fix_json(self):
        def merge_statuses(unmerged_status):
            merged_status: list[dict[str, Any]] = [
                val for lst in unmerged_status.values() for val in lst
            ]
            return merged_status

        for metadata_filepath in self.metadatas:
            if (
                metadata_filepath.suffix != ".json"
                or "__legacy__" in metadata_filepath.parts
            ):
                continue
            final_content_type = None
            archive = False
            new_metadata_set = main_helper.import_json(metadata_filepath)
            content_types = self.subscription.get_api().ContentTypes().get_keys()
            final_stem = metadata_filepath.stem
            if final_stem[-1].isdigit():
                final_stem = final_stem.removesuffix(f"_{final_stem[-1]}")
            if final_stem in content_types:
                final_content_type = final_stem
            else:
                for item in content_types:
                    if item in metadata_filepath.parts:
                        final_content_type = item
                        break
                if final_stem == "Archived":
                    archive = True
                    final_content_type = "Posts"
                    pass
                if not final_content_type:
                    # If we get an key error here, we need to move the json file to correct folder or we can resolve it by getting content_type by looking at the directory path
                    # directory_set = set()
                    # merged = merge_statuses(new_metadata_set)
                    # for item in merged:
                    #     directory_set.add(item["directory"])
                    final_content_type = new_metadata_set["content_type"]
            final_metadata_set = {}
            content_json = {}
            if isinstance(new_metadata_set, list):
                if len(new_metadata_set) > 1:
                    pass
                for temp_set in new_metadata_set:
                    content_json[temp_set["type"]] = {
                        "valid": temp_set["valid"],
                        "invalid": temp_set["invalid"],
                    }
                final_metadata_set["version"] = 1.8
                final_metadata_set["content"] = content_json
                pass
            elif (
                "type" not in new_metadata_set
                and "version" not in new_metadata_set
                and "valid" in new_metadata_set
            ):
                content_json[metadata_filepath.stem] = {
                    "valid": new_metadata_set["valid"],
                    "invalid": new_metadata_set["invalid"],
                }
                final_metadata_set["version"] = 1.9
                final_metadata_set["content"] = content_json
                pass
            elif "type" in new_metadata_set:
                # Usually means there's another json file we'd have to merge into this, like Images.json and Videos.json.
                temp_content_json = {
                    "version": 2.0,
                    "content_type": final_content_type,
                    "content": {},
                }
                temp_content_json["content"][new_metadata_set["type"]] = {
                    "valid": new_metadata_set["valid"],
                    "invalid": new_metadata_set["invalid"],
                }
                content_json = temp_content_json["content"]
                final_metadata_set = temp_content_json
                pass
            else:
                content_json = new_metadata_set.get("content", new_metadata_set)
            if final_metadata_set:
                final_metadata_set["content_type"] = final_content_type
                main_helper.export_json(final_metadata_set, metadata_filepath)
            for media_type, status in content_json.items():
                merged_status = merge_statuses(status)
                if merged_status:
                    for item_json in merged_status:
                        if not final_content_type:
                            raise Exception("content type not found")

                        def assign_metadata(
                            content_metadata: ContentMetadata, item_json: dict[str, Any]
                        ):
                            content_metadata.__legacy__ = True
                            content_metadata.created_at_string = item_json["postedAt"]
                            content_metadata.paid = item_json.get("paid", False)
                            content_metadata.price = item_json.get("price", 0)
                            content_metadata.text = item_json["text"]
                            content_metadata.user_id = self.subscription.id
                            content_metadata.archived = archive
                            for asset_json in item_json.get("medias", [item_json]):
                                urls = asset_json.get("links") or []
                                if "link" in asset_json:
                                    urls.append(asset_json["link"])

                                media_metadata = content_metadata.find_media(
                                    asset_json.get("media_id"), urls
                                )
                                if not media_metadata:
                                    media_metadata = MediaMetadata(
                                        asset_json.get("media_id"),
                                        media_type,
                                        content_metadata,
                                        urls,
                                    )
                                    pass
                                media_metadata.directory = Path(asset_json["directory"])
                                media_metadata.filename = asset_json["filename"]
                                media_metadata.size = asset_json.get("size", 0)
                                content_metadata.medias.append(media_metadata)

                        content_metadata = None
                        if isinstance(item_json, list):
                            for item_item_json in item_json:
                                content_metadata = self.find_content_metadata_by_id(
                                    item_item_json["post_id"],
                                    content_metadats=self.legacy_content_metadatas,
                                )
                                if not content_metadata:
                                    content_metadata = ContentMetadata(
                                        item_item_json["post_id"], final_content_type
                                    )
                                    self.legacy_content_metadatas.append(
                                        content_metadata
                                    )
                                    pass
                                assign_metadata(content_metadata, item_item_json)
                                pass
                            pass
                        else:
                            content_metadata = self.find_content_metadata_by_id(
                                item_json["post_id"],
                                content_metadats=self.legacy_content_metadatas,
                            )
                            if not content_metadata:
                                content_metadata = ContentMetadata(
                                    item_json["post_id"], final_content_type
                                )
                                self.legacy_content_metadatas.append(content_metadata)
                            assign_metadata(content_metadata, item_json)

        return self.legacy_content_metadatas

    def fix_sqlite(self):
        subscription = self.subscription
        if not self.db_manager:
            raise Exception("DatabaseManager has not been assigned")

        api = subscription.get_api()
        site_settings = api.get_site_settings()
        config = api.config
        if not (config and site_settings):
            return
        # We could put this in process_legacy_metadata
        legacy_db_manager = DatabaseManager().get_sqlite_db(
            legacy_metadata_path, legacy=True
        )

        final_result = legacy_db_manager.legacy_sqlite_updater(api_type, subscription)
        for metadata in final_result:
            content_metadata = ContentMetadata(metadata.post_id, api_type)
            augh = content_metadata.resolve_extractor(MetadataExtractor(metadata))
            pass
        new_metadata_object.extend(final_result)
        return new_metadata_object

    def delete_metadatas(self):
        for legacy_metadata in self.redundant_metadatas:
            if self.subscription.get_api().get_site_settings().delete_legacy_metadata:
                legacy_metadata.unlink()
            else:
                if legacy_metadata.exists():
                    new_filepath = Path(
                        legacy_metadata.parent,
                        "__legacy__",
                        legacy_metadata.name,
                    )
                    new_filepath.parent.mkdir(exist_ok=True)
                    legacy_metadata.rename(new_filepath)

    async def fix_archived_db(
        self,
    ):
        api = self.subscription.get_api()
        directory_manager = self.filesystem_manager.get_directory_manager(
            self.subscription.id
        )
        for final_metadata in directory_manager.user.legacy_metadata_directories:
            archived_database_path = final_metadata.joinpath("Archived.db")
            if archived_database_path.exists():
                archived_db_manager = DatabaseManager().get_sqlite_db(
                    archived_database_path, legacy=True
                )
                if not archived_db_manager.session.bind:
                    continue
                for api_type, _value in api.ContentTypes():
                    database_path = final_metadata.joinpath(f"{api_type}.db")
                    legacy_db_manager = DatabaseManager().get_sqlite_db(
                        database_path, legacy=True
                    )
                    database_name = api_type.lower()
                    result: bool = inspect(
                        archived_db_manager.session.bind.engine
                    ).has_table(database_name)
                    if result:
                        archived_db_manager.alembic_directory = (
                            archived_db_manager.alembic_directory.parent.with_name(
                                database_name
                            ).joinpath("alembic")
                        )
                        archived_db_manager.run_migrations(legacy=True)
                        legacy_db_manager.run_migrations(legacy=True)

                        db_manager = DatabaseManager().get_sqlite_db(database_path)
                        modern_database_session = db_manager.session
                        db_collection = DBCollection()
                        database = db_collection.database_picker("user_data")
                        if not database:
                            return
                        table_name = database.table_picker(api_type, True)
                        if not table_name:
                            return
                        archived_result = archived_db_manager.session.query(
                            table_name
                        ).all()
                        for item in archived_result:
                            result2 = (
                                modern_database_session.query(table_name)
                                .filter(table_name.post_id == item.post_id)
                                .first()
                            )
                            if not result2:
                                item2 = item.__dict__
                                item2.pop("id")
                                item2.pop("_sa_instance_state")
                                item = table_name(**item2)
                                item.archived = True
                                modern_database_session.add(item)
                            else:
                                result2.archived = True
                        modern_database_session.commit()
                        modern_database_session.close()
                archived_db_manager.session.commit()
                archived_db_manager.session.close()
                new_filepath = Path(
                    archived_database_path.parent,
                    "__legacy__",
                    archived_database_path.name,
                )
                new_filepath.parent.mkdir(exist_ok=True)
                archived_database_path.rename(new_filepath)

    def export(self, api_type: str, datas: list[dict[str, Any]]):
        if api_type == "Posts":
            self.db_manager
        pass

    def find_content_metadata_by_id(
        self, id: int, content_metadats: list[ContentMetadata] = []
    ):
        content_metadats = content_metadats or self.content_metadatas
        for content in content_metadats:
            if content.content_id == id:
                return content

    def update_metadata_filepaths(self):
        filesystem_manager = self.filesystem_manager.get_file_manager(
            self.subscription.id
        )

        directory_manager = filesystem_manager.directory_manager
        if directory_manager:
            for i, metadata_filepath in enumerate(self.metadatas):
                new_m_f = directory_manager.user.metadata_directory.joinpath(
                    metadata_filepath.name
                )
                if metadata_filepath != new_m_f:
                    if (
                        "__legacy__" in metadata_filepath.parts
                        or ".db" != metadata_filepath.suffix
                    ):
                        continue
                    counter = 0
                    while True:
                        if not new_m_f.exists():
                            # If there's metadata present already before the directory is created, we'll create it here
                            directory_manager.user.metadata_directory.mkdir(
                                exist_ok=True, parents=True
                            )
                            self.filesystem_manager.move(metadata_filepath, new_m_f)
                            break
                        else:
                            new_m_f = new_m_f.with_stem(
                                f"{metadata_filepath.stem}_{counter}"
                            )
                            counter += 1
                self.metadatas[i] = new_m_f



from typing import Any


class OptionManager:
    def __init__(self) -> None:
        self.performer_options: OptionsFormat | None = None
        self.subscription_options: OptionsFormat | None = None
        pass

    async def create_option(
        self,
        items: list[Any],
        options_type: str,
        auto_choice: list[int | str] | int | str | bool = False,
    ):
        option = await OptionsFormat(items, options_type, auto_choice).formatter()
        return option


class OptionsFormat:
    def __init__(
        self,
        items: list[Any],
        options_type: str,
        auto_choice: list[int | str] | int | str | bool = False,
    ) -> None:
        self.items = items
        self.item_keys: list[str] = []
        self.string = ""
        self.options_type = options_type
        self.auto_choice = auto_choice
        self.final_choices = []

    async def formatter(self):
        options_type = self.options_type
        final_string = f"Choose {options_type.capitalize()}: 0 = All"
        if type(self.auto_choice) == int:
            self.auto_choice = str(self.auto_choice)

        if isinstance(self.auto_choice, str):
            self.auto_choice = [x for x in self.auto_choice.split(",") if x]
            self.auto_choice = (
                True
                if any(x in ["0", "all"] for x in self.auto_choice)
                else self.auto_choice
            )

        if isinstance(self.auto_choice, list):
            self.auto_choice = [x for x in self.auto_choice if x]

        match options_type:
            case "sites":
                self.item_keys = self.items
                my_string = " | ".join(
                    map(lambda x: f"{self.items.index(x)+1} = {x}", self.items)
                )
                final_string = f"{final_string} | {my_string}"
                self.string = final_string
                final_list = await self.choose_option()
                self.final_choices = [
                    key
                    for choice in final_list
                    for key in self.items
                    if choice.lower() == key.lower()
                ]
            case "profiles":
                self.item_keys = [x.auth_details.username for x in self.items]
                my_string = " | ".join(
                    map(
                        lambda x: f"{self.items.index(x)+1} = {x.auth_details.username}",
                        self.items,
                    )
                )
                final_string = f"{final_string} | {my_string}"
                self.string = final_string
                final_list = await self.choose_option()
                self.final_choices = [
                    key
                    for choice in final_list
                    for key in self.items
                    if choice.lower() == key.auth_details.username.lower()
                ]
                set1 = set(self.final_choices)
                set2 = set(self.items)
                difference = list(set2 - set1)
                for auth in difference:
                    await auth.session_manager.active_session.close()
            case "subscriptions":
                subscription_users = [x for x in self.items]
                self.item_keys = [x.username for x in subscription_users]
                my_string = " | ".join(
                    map(
                        lambda x: f"{subscription_users.index(x)+1} = {x.username}",
                        subscription_users,
                    )
                )
                final_string = f"{final_string} | {my_string}"
                self.string = final_string
                final_list = await self.choose_option()
                self.final_choices = [
                    key
                    for choice in final_list
                    for key in subscription_users
                    if choice.lower() == key.username.lower()
                ]

            case "contents":
                self.item_keys = self.items
                my_string = " | ".join(
                    map(lambda x: f"{self.items.index(x)+1} = {x}", self.items)
                )
                final_string = f"{final_string} | {my_string}"
                self.string = final_string
                final_list = await self.choose_option()
                self.final_choices = [
                    key
                    for choice in final_list
                    for key in self.items
                    if choice.lower() == key.lower()
                ]
            case "medias":
                self.item_keys = self.items
                my_string = " | ".join(
                    map(lambda x: f"{self.items.index(x)+1} = {x}", self.items)
                )
                final_string = f"{final_string} | {my_string}"
                self.string = final_string
                final_list = await self.choose_option()
                self.final_choices = [
                    key
                    for choice in final_list
                    for key in self.items
                    if choice.lower() == key.lower()
                ]
            case _:
                final_list = []
        return self

    async def choose_option(self):
        def process_option(input_values: list[str]):
            input_list_2: list[str] = []
            for input_value in input_values:
                if input_value.isdigit():
                    try:
                        input_list_2.append(self.item_keys[int(input_value) - 1])
                    except IndexError:
                        continue
                else:
                    x = [x for x in self.item_keys if x.lower() == input_value.lower()]
                    input_list_2.extend(x)
            return input_list_2

        input_list: list[str] = [x.lower() for x in self.item_keys]
        final_list: list[str] = []
        if self.auto_choice:
            if not self.scrape_all():
                if isinstance(self.auto_choice, list):
                    input_values = [str(x).lower() for x in self.auto_choice]
                    input_list = process_option(input_values)
        else:
            print(self.string)
            input_value = input().lower()
            if input_value != "0" and input_value != "all":
                input_values = input_value.split(",")
                input_list = process_option(input_values)
        final_list = input_list
        return final_list

    def scrape_all(self):
        status = False
        if (
            self.auto_choice == True
            or isinstance(self.auto_choice, list)
            and isinstance(self.auto_choice[0], str)
            and (
                self.auto_choice[0].lower() == "all"
                or self.auto_choice[0].lower() == "0"
            )
        ):
            status = True
        return status

    def return_auto_choice(self):
        identifiers = []
        if isinstance(self.auto_choice, list):
            identifiers = [x for x in self.auto_choice if not isinstance(x, bool)]
        return identifiers



from typing import Optional, cast

import sqlalchemy
from sqlalchemy.orm.decl_api import declarative_base

from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.api_model import (
    ApiModel,
)
from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.media_model import (
    TemplateMediaModel,
)

Base = declarative_base()
LegacyBase = declarative_base()


class profiles_table(Base):
    __tablename__ = "profiles"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    user_id = cast(int, sqlalchemy.Column(sqlalchemy.Integer, nullable=False))
    username = sqlalchemy.Column(sqlalchemy.String, unique=True, nullable=False)


class stories_table(ApiModel, Base):
    ApiModel.__tablename__ = "stories"


class posts_table(ApiModel, Base):
    ApiModel.__tablename__ = "posts"


class messages_table(ApiModel, Base):
    ApiModel.__tablename__ = "messages"
    user_id = cast(Optional[int], sqlalchemy.Column(sqlalchemy.Integer))

    class api_legacy_table(ApiModel, LegacyBase):
        pass


class products_table(ApiModel, Base):
    ApiModel.__tablename__ = "products"
    title = sqlalchemy.Column(sqlalchemy.String)


class others_table(ApiModel, Base):
    ApiModel.__tablename__ = "others"


# class comments_table(api_table,Base):
#     api_table.__tablename__ = "comments"


class media_table(TemplateMediaModel, Base):
    class media_legacy_table(TemplateMediaModel().legacy_2(LegacyBase), LegacyBase):
        pass


def table_picker(table_name: str, legacy: bool = False):
    match table_name:
        case "Stories" | "Highlights":
            table = stories_table
        case "Posts":
            table = posts_table
        case "Messages" | "Chats" | "MassMessages":
            table = messages_table if not legacy else messages_table().api_legacy_table
        case "Products":
            table = products_table
        case "Others":
            table = others_table
        case _:
            raise Exception(f'"{table_name}" is an invalid table name')
    return table



from pathlib import Path
from typing import Any, Union

from ultima_scraper_api.apis.onlyfans.classes.auth_model import create_auth
from ultima_scraper_api.apis.onlyfans.classes.hightlight_model import create_highlight
from ultima_scraper_api.apis.onlyfans.classes.message_model import create_message
from ultima_scraper_api.apis.onlyfans.classes.post_model import create_post
from ultima_scraper_api.apis.onlyfans.classes.story_model import create_story
from ultima_scraper_api.apis.onlyfans.classes.user_model import create_user
from ultima_scraper_api.apis.onlyfans.onlyfans import OnlyFansAPI
from ultima_scraper_renamer.reformat import ReformatManager

from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    ContentMetadata,
)
from ultima_scraper_collection.managers.option_manager import OptionManager
from ultima_scraper_collection.modules.module_streamliner import StreamlinedDatascraper


class OnlyFansDataScraper(StreamlinedDatascraper):
    def __init__(self, api: OnlyFansAPI, option_manager: OptionManager) -> None:
        self.api = api
        self.option_manager = option_manager
        StreamlinedDatascraper.__init__(self, self)

    # Scrapes the API for content
    async def media_scraper(
        self,
        post_result: Union[create_story, create_post, create_message],
        subscription: create_user,
        formatted_directory: Path,
        api_type: str,
    ) -> dict[str, Any]:
        authed = subscription.get_authed()
        api = authed.api
        site_settings = api.get_site_settings()
        if not site_settings:
            return {}
        new_set: dict[str, Any] = {"content": []}
        directories: list[Path] = []
        if api_type == "Stories":
            pass
        if api_type == "Archived":
            pass
        if api_type == "Posts":
            pass
        if api_type == "Messages":
            pass
        from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
            Extractor,
        )

        content_metadata = ContentMetadata(post_result.id, api_type)
        content_metadata.resolve_extractor(Extractor(post_result))
        matches = [
            s for s in site_settings.ignored_keywords if s in content_metadata.text
        ]
        if matches:
            print("Ignoring - ", f"PostID: {content_metadata.content_id}")
            return {}
        for asset in content_metadata.medias:
            if asset.urls:
                reformat_manager = ReformatManager(authed, self.filesystem_manager)
                reformat_item = reformat_manager.prepare_reformat(asset)
                file_directory = reformat_item.reformat(
                    site_settings.file_directory_format
                )
                reformat_item.directory = file_directory
                file_path = reformat_item.reformat(site_settings.filename_format)
                asset.directory = file_directory
                asset.filename = file_path.name

                if file_directory not in directories:
                    directories.append(file_directory)
        new_set["content"].append(content_metadata)
        new_set["directories"] = directories
        return new_set

    async def get_all_stories(self, subscription: create_user):
        """
        get_all_stories(subscription: create_user)

        This function returns a list of all stories and highlights from the given subscription.

        Arguments:
        subscription (create_user): An instance of the create_user class.

        Returns:
        list[create_highlight | create_story]: A list containing all stories and highlights from the subscription.
        """
        master_set: list[create_highlight | create_story] = []
        master_set.extend(await subscription.get_stories())
        master_set.extend(await subscription.get_archived_stories())
        highlights = await subscription.get_highlights()
        valid_highlights: list[create_highlight | create_story] = []
        for highlight in highlights:
            resolved_highlight = await subscription.get_highlights(
                hightlight_id=highlight.id
            )
            valid_highlights.extend(resolved_highlight)
        master_set.extend(valid_highlights)
        return master_set

    async def get_all_posts(self, subscription: create_user):
        temp_master_set = await subscription.get_posts()
        # get_archived_posts uses normal posts
        await subscription.get_archived_posts()
        return temp_master_set

    async def get_all_subscriptions(
        self,
        authed: create_auth,
        identifiers: list[int | str] = [],
        refresh: bool = True,
    ):
        """
        get_all_subscriptions(authed: create_auth, identifiers: list[int | str] = [], refresh: bool = True)

        This function returns a list of all subscriptions from the given authenticated user.

        Arguments:
        authed (create_auth): An instance of the create_auth class.
        identifiers (list[int | str], optional): A list of identifiers (username or id) for the subscriptions. Defaults to an empty list.
        refresh (bool, optional): A flag indicating whether to refresh the list of subscriptions. Defaults to True.

        Returns:
        list[create_subscription]: A list of all subscriptions, sorted by expiredAt, from the authenticated user.
        """
        subscriptions = await authed.get_subscriptions(
            identifiers=identifiers, refresh=refresh, sub_type="active"
        )
        subscriptions.sort(key=lambda x: x.subscribed_by_expire_date)
        return subscriptions



# type: ignore
"""content

Revision ID: 5b4bea08c27f
Revises: 194e05269f09
Create Date: 2021-02-04 02:59:05.010106

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5b4bea08c27f"
down_revision = "194e05269f09"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.add_column(sa.Column("preview", sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.drop_column("preview")

    # ### end Alembic commands ###



# Belongs to onlyfans.py (OnlyFansDataScraper Class)
# async def process_mass_messages(
#     self, authed: create_auth, mass_messages: list[create_message]
# ):
#     def compare_message(queue_id, remote_messages):
#         for message in remote_messages:
#             if "isFromQueue" in message and message["isFromQueue"]:
#                 if queue_id == message["queueId"]:
#                     return message
#                 print
#         print

#     global_found = []
#     chats = []
#     api = authed.get_api()
#     site_settings = api.get_site_settings()
#     config = api.config
#     if not (config and site_settings):
#         return
#     settings = config.settings
#     salt = settings.random_string
#     encoded = f"{salt}"
#     encoded = encoded.encode("utf-8")
#     hash = hashlib.md5(encoded).hexdigest()
#     profile_directory = authed.directory_manager.profile.metadata_directory
#     mass_message_path = profile_directory.joinpath("Mass Messages.json")
#     chats_path = profile_directory.joinpath("Chats.json")
#     if os.path.exists(chats_path):
#         chats = main_helper.import_json(chats_path)
#     date_object = datetime.today()
#     date_string = date_object.strftime("%d-%m-%Y %H:%M:%S")
#     for mass_message in mass_messages:
#         if "status" not in mass_message:
#             mass_message["status"] = ""
#         if "found" not in mass_message:
#             mass_message["found"] = {}
#         if "hashed_ip" not in mass_message:
#             mass_message["hashed_ip"] = ""
#         mass_message["hashed_ip"] = mass_message.get("hashed_ip", hash)
#         mass_message["date_hashed"] = mass_message.get("date_hashed", date_string)
#         if mass_message["isCanceled"]:
#             continue
#         queue_id = mass_message["id"]
#         text = mass_message["textCropped"]
#         text = html.unescape(text)
#         mass_found = mass_message["found"]
#         media_type = mass_message.get("mediaType")
#         media_types = mass_message.get("mediaTypes")
#         if mass_found or (not media_type and not media_types):
#             continue
#         identifier = None
#         if chats:
#             list_chats = chats
#             for chat in list_chats:
#                 identifier = chat["identifier"]
#                 messages = chat["messages"]["list"]
#                 mass_found = compare_message(queue_id, messages)
#                 if mass_found:
#                     mass_message["found"] = mass_found
#                     mass_message["status"] = True
#                     break
#         if not mass_found:
#             list_chats = authed.search_messages(text=text, limit=2)
#             if not list_chats:
#                 continue
#             for item in list_chats["list"]:
#                 user = item["withUser"]
#                 identifier = user["id"]
#                 messages = []
#                 print("Getting Messages")
#                 keep = ["id", "username"]
#                 list_chats2 = [x for x in chats if x["identifier"] == identifier]
#                 if list_chats2:
#                     chat2 = list_chats2[0]
#                     messages = chat2["messages"]["list"]
#                     messages = authed.get_messages(
#                         identifier=identifier, resume=messages
#                     )
#                     for message in messages:
#                         message["withUser"] = {k: item["withUser"][k] for k in keep}
#                         message["fromUser"] = {
#                             k: message["fromUser"][k] for k in keep
#                         }
#                     mass_found = compare_message(queue_id, messages)
#                     if mass_found:
#                         mass_message["found"] = mass_found
#                         mass_message["status"] = True
#                         break
#                 else:
#                     item2 = {}
#                     item2["identifier"] = identifier
#                     item2["messages"] = authed.get_messages(identifier=identifier)
#                     chats.append(item2)
#                     messages = item2["messages"]["list"]
#                     for message in messages:
#                         message["withUser"] = {k: item["withUser"][k] for k in keep}
#                         message["fromUser"] = {
#                             k: message["fromUser"][k] for k in keep
#                         }
#                     mass_found = compare_message(queue_id, messages)
#                     if mass_found:
#                         mass_message["found"] = mass_found
#                         mass_message["status"] = True
#                         break
#                     print
#                 print
#             print
#         if not mass_found:
#             mass_message["status"] = False
#     main_helper.export_json(chats, chats_path)
#     for mass_message in mass_messages:
#         found = mass_message["found"]
#         if found and found["media"]:
#             user = found["withUser"]
#             identifier = user["id"]
#             print
#             date_hashed_object = datetime.strptime(
#                 mass_message["date_hashed"], "%d-%m-%Y %H:%M:%S"
#             )
#             next_date_object = date_hashed_object + timedelta(days=1)
#             print
#             if mass_message["hashed_ip"] != hash or date_object > next_date_object:
#                 print("Getting Message By ID")
#                 x = await authed.get_message_by_id(
#                     identifier=identifier, identifier2=found["id"], limit=1
#                 )
#                 new_found = x["result"]["list"][0]
#                 new_found["withUser"] = found["withUser"]
#                 mass_message["found"] = new_found
#                 mass_message["hashed_ip"] = hash
#                 mass_message["date_hashed"] = date_string
#             global_found.append(found)
#         print
#     print
#     main_helper.export_json(mass_messages, mass_message_path)
#     return global_found



from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

import ultima_scraper_api
from alembic import command
from alembic.config import Config
from alembic.migration import MigrationContext
from sqlalchemy import create_engine, func
from sqlalchemy.orm import DeclarativeBase, scoped_session, sessionmaker

from ultima_scraper_collection.managers.database_manager.connections.sqlite.models import (
    user_database,
)

if TYPE_CHECKING:

    from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
        ContentMetadata,
    )

user_types = ultima_scraper_api.user_types


class DBCollection(object):
    def __init__(self) -> None:
        self.user_database = user_database

    def database_picker(self, database_name: str):
        match database_name:
            case "user_data":
                database = self.user_database
            case _:
                raise Exception(f'"{database_name}" is an invalid database name')
        return database


class SqliteDatabase(DeclarativeBase):
    def __init__(self) -> None:
        self.name = Path()
        self.info = ""
        self.session_factory = sessionmaker()
        self.session = self.create_session()
        self.root_directory = Path(__file__).parent
        self.alembic_directory = Path()
        self.migration_directory = Path()

    def init_db(self, name: Path, legacy: bool = False):
        self.name: Path = name
        self.info = f"sqlite:///{name}"
        self.session_factory = sessionmaker(bind=self.create_engine(), autocommit=False)
        self.session = self.create_session()
        self.alembic_directory = Path(__file__).parent.joinpath(
            f"{'databases' if not legacy else 'legacy_databases'}",
            self.name.stem.lower(),
            "alembic",
        )
        self.migration_directory = self.alembic_directory.parent.joinpath("migration")
        return self

    def create_engine(self):
        return create_engine(
            self.info,
            connect_args={},
        )

    def create_session(self):
        return scoped_session(self.session_factory)

    def execute(self, statement: Any):
        result = self.session.execute(statement)
        return result

    def generate_migration(self):

        if not self.session.bind:
            return
        conn = self.session.bind.engine.connect()
        context = MigrationContext.configure(conn)
        current_rev = context.get_current_revision()
        alembic_cfg = Config(self.migration_directory.joinpath("alembic.ini"))
        alembic_cfg.set_main_option(
            "script_location",
            self.migration_directory.joinpath("alembic").as_posix(),
        )
        alembic_cfg.set_main_option("sqlalchemy.url", self.info)
        if not current_rev:
            _ggg = command.revision(alembic_cfg, autogenerate=True)
        else:
            _ggg = command.revision(alembic_cfg, autogenerate=True, head=current_rev)
        self.run_migrations()
        return True

    def run_migrations(self, legacy: bool = False) -> None:
        while True:
            try:

                migration_directory = (
                    self.alembic_directory.parent.joinpath("migration")
                    if legacy
                    else self.migration_directory
                )

                alembic_cfg = Config(migration_directory.joinpath("alembic.ini"))
                alembic_cfg.set_main_option(
                    "script_location",
                    migration_directory.joinpath("alembic").as_posix(),
                )
                alembic_cfg.set_main_option("sqlalchemy.url", self.info)
                command.upgrade(alembic_cfg, "head")
                break
            except Exception as e:
                print(e)
                pass

    def revert_migration(self):
        while True:
            try:
                alembic_cfg = Config(self.migration_directory.joinpath("alembic.ini"))
                alembic_cfg.set_main_option(
                    "script_location",
                    self.migration_directory.joinpath("alembic").as_posix(),
                )
                alembic_cfg.set_main_option("sqlalchemy.url", self.info)
                command.downgrade(alembic_cfg, "-1")
                break
            except Exception as e:
                print(e)
                pass

    def import_metadata(
        self, datas: list["ContentMetadata"], api_type: str | None = None
    ):
        database_path = self.name
        database_path.parent.mkdir(parents=True, exist_ok=True)
        self.run_migrations()
        db_collection = DBCollection()
        database = db_collection.database_picker(database_path.stem)
        database_session = self.session
        for post in datas:
            if post.api_type:
                api_type = post.api_type
            api_table = database.table_picker(api_type)
            if not api_table:
                return
            post_id = post.content_id
            created_at_string = post.created_at_string
            date_object = None
            if created_at_string:
                date_object = datetime.strptime(created_at_string, "%d-%m-%Y %H:%M:%S")
            result = database_session.query(api_table)
            post_db = result.filter_by(post_id=post_id).first()
            if not post_db:
                post_db = api_table()
            else:
                pass
            if api_type == "Messages":
                post_db.user_id = post.user_id
            post_db.post_id = post_id
            post_db.text = post.text
            post_db.price = post.price
            post_db.paid = post.paid
            post_db.archived = post.archived
            if date_object:
                post_db.created_at = date_object
            database_session.add(post_db)
            for media in post.medias:
                if media.media_type == "Texts":
                    continue
                created_at = media.created_at
                if not isinstance(created_at, datetime):
                    date_object = datetime.strptime(created_at, "%d-%m-%Y %H:%M:%S")
                media_id = media.id
                result = database_session.query(database.media_table)
                media_db = result.filter_by(post_id=post_id, media_id=media_id).first()
                if not media_db:
                    media_db = result.filter_by(
                        filename=media.filename, created_at=date_object
                    ).first()
                    if not media_db:
                        media_db = database.media_table()
                else:
                    pass
                if (
                    post.__legacy__
                    and media_db.media_id != media.id
                    and media_db.media_id
                ):
                    media_id = media_db.media_id

                media_db.media_id = media_id
                media_db.post_id = post_id
                media_db.size = media.size if media_db.size is None else media_db.size
                media_db.link = media.urls[0] if media.urls else None
                media_db.preview = media.preview
                media_db.directory = (
                    media.directory.as_posix() if media.directory else None
                )
                media_db.filename = media.filename
                media_db.api_type = api_type
                media_db.media_type = media.media_type
                media_db.linked = media.linked
                if date_object:
                    media_db.created_at = date_object
                database_session.add(media_db)
        database_session.commit()
        database_session.close()
        return True

    def legacy_sqlite_updater(
        self,
        api_type: str,
        subscription: user_types,
    ):
        final_result: list[dict[str, Any]] = []
        legacy_metadata_path = self.name
        if legacy_metadata_path.exists():
            self.run_migrations(legacy=True)
            database_name = "user_data"
            database_session = self.session
            db_collection = DBCollection()
            database = db_collection.database_picker(database_name)
            if database:
                if api_type == "Messages":
                    api_table_table = database.table_picker(api_type, True)
                else:
                    api_table_table = database.table_picker(api_type)
                media_table_table = database.media_table.media_legacy_table
                if api_table_table:
                    result = database_session.query(api_table_table).all()
                    result2 = database_session.query(media_table_table).all()
                    for item in result:
                        for item2 in result2:
                            if item.post_id != item2.post_id:
                                continue
                            item.medias.append(item2)
                        item.user_id = subscription.id
                        final_result.append(item)
            database_session.close()
        return final_result

    def find_table(self, name: str):
        table = [x for x in self.metadata.sorted_tables if x.name == name]
        if table:
            return table[0]

    def get_count(self, q: Any):
        count_q = q.statement.with_only_columns(func.count()).order_by(None)
        count: int = q.session.execute(count_q).scalar()
        return count



import ultima_scraper_api
from ultima_scraper_api.apis.onlyfans import onlyfans
from ultima_scraper_collection import datascraper_types
from ultima_scraper_collection.managers.datascraper_manager.datascrapers.fansly import (
    FanslyDataScraper,
)
from ultima_scraper_collection.managers.datascraper_manager.datascrapers.onlyfans import (
    OnlyFansDataScraper,
)
from ultima_scraper_collection.managers.option_manager import OptionManager


class DataScraperManager:
    def __init__(self) -> None:
        self.active_datascraper = None
        self.datascrapers: list[datascraper_types] = []

    def select_datascraper(
        self,
        api: ultima_scraper_api.api_types,
        option_manager: OptionManager = OptionManager(),
    ):
        self.active_datascraper = self.get_datascraper(api)
        if not self.active_datascraper:
            self.active_datascraper = self.add_datascraper(api, option_manager)
        return self.active_datascraper

    def add_datascraper(
        self, api: ultima_scraper_api.api_types, option_manager: OptionManager
    ):
        if isinstance(api, onlyfans.OnlyFansAPI):
            datascraper = OnlyFansDataScraper(api, option_manager)
        else:
            datascraper = FanslyDataScraper(api, option_manager)
        self.datascrapers.append(datascraper)
        return datascraper

    def get_datascraper(self, api: ultima_scraper_api.api_types):
        for datascraper in self.datascrapers:
            if datascraper.api.site_name == api.site_name:
                return datascraper
        return None



### api_table.py ###

from datetime import datetime
from typing import Any, cast

import sqlalchemy


class TemplateMediaModel:
    __tablename__ = "medias"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    media_id = sqlalchemy.Column(sqlalchemy.Integer)
    post_id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)
    link = cast(str, sqlalchemy.Column(sqlalchemy.String))
    directory = cast(str, sqlalchemy.Column(sqlalchemy.String))
    filename = cast(str, sqlalchemy.Column(sqlalchemy.String))
    size = cast(int | None, sqlalchemy.Column(sqlalchemy.Integer, default=0))
    api_type = cast(str, sqlalchemy.Column(sqlalchemy.String))
    media_type = sqlalchemy.Column(sqlalchemy.String)
    preview = sqlalchemy.Column(sqlalchemy.Integer, default=0)
    linked = sqlalchemy.Column(sqlalchemy.String, default=None)
    downloaded = cast(bool, sqlalchemy.Column(sqlalchemy.Integer, default=0))
    created_at = cast(datetime, sqlalchemy.Column(sqlalchemy.TIMESTAMP))

    def legacy(self, Base: Any):
        class legacy_media_table(Base):
            __tablename__ = "medias"
            id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
            post_id = sqlalchemy.Column(sqlalchemy.Integer)
            link = sqlalchemy.Column(sqlalchemy.String)
            directory = sqlalchemy.Column(sqlalchemy.String)
            filename = sqlalchemy.Column(sqlalchemy.String)
            size = sqlalchemy.Column(sqlalchemy.Integer, default=None)
            media_type = sqlalchemy.Column(sqlalchemy.String)
            downloaded = sqlalchemy.Column(sqlalchemy.Integer, default=0)
            created_at = sqlalchemy.Column(sqlalchemy.DATETIME)

        return legacy_media_table

    def legacy_2(self, Base: Any):
        class legacy_media_table(Base):
            __tablename__ = "medias"
            id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
            media_id = sqlalchemy.Column(sqlalchemy.Integer, unique=True)
            post_id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)
            link = cast(str, sqlalchemy.Column(sqlalchemy.String))
            directory = cast(str, sqlalchemy.Column(sqlalchemy.String))
            filename = cast(str, sqlalchemy.Column(sqlalchemy.String))
            size = cast(int, sqlalchemy.Column(sqlalchemy.Integer, default=None))
            media_type = sqlalchemy.Column(sqlalchemy.String)
            preview = sqlalchemy.Column(sqlalchemy.Integer, default=0)
            linked = sqlalchemy.Column(sqlalchemy.String, default=None)
            downloaded = cast(bool, sqlalchemy.Column(sqlalchemy.Integer, default=0))
            created_at = cast(datetime, sqlalchemy.Column(sqlalchemy.TIMESTAMP))

        return legacy_media_table
